// instrumentation core unit of hinstrumentation library;
unit instrCoreU;

interface

uses
  {$ifNDef FPC}
  Windows,
  {$endIf}
  types,
  sysutils,
  Classes,
  EpikTimer;

procedure instration_begin(const aUnit, aProc: string; aLine: Integer);
procedure instration_end(const aUnit, aProc: string; aLine: Integer);

const
  MaxDepth = 100;
  MaxThreads = 100;
  {$ifNDef FPC}
  LineEnding = sLineBreak;
  {$endIf}

type
  TTimerDataDynArray = array of TimerData;

  {$ifNDef FPC}
  TThreadId = DWORD;
  {$endIf}

  { TThreadInfo }

  TThreadInfo = class
  private
    TimerStack: TTimerDataDynArray;
    Timer: TEpikTimer;
    Level: Integer;
    FID: TThreadID;
    Log: TStream;
    procedure PrepareTimerStack;
    procedure WriteLog(s: string);
    procedure PushTimer;
    function PopTimer: Extended;
    function GetIndent: string;
    function GetTimeStr: string;
  public
    property ID: TThreadID read FID write FID;
    constructor Create(aID: TThreadID);
    procedure Begin_(const aUnit, aProc: string; aLine: Integer);
    procedure End_(const aUnit, aProc: string; aLine: Integer);
    destructor Destroy; override;
  end;

  TThreadInfoDynArray = array of TThreadInfo;

var
  ThreadsLocker: TRTLCriticalSection;
  Threads: TThreadInfoDynArray;

function GetLogFileName: string;

implementation

{$ifNDef FPC}
procedure InitCriticalSection(var a: TRTLCriticalSection);
begin
  InitializeCriticalSection(a);
end;

procedure DoneCriticalSection(var a: TRTLCriticalSection);
begin
  DeleteCriticalSection(a);
end;
{$endIf}

function FindThreadInfoById(aId: TThreadId; aThreads: TThreadInfoDynArray): TThreadInfo;
var
  i: Integer;
  thread: TThreadInfo;
begin
  result := nil;
  for i := 0 to Length(aThreads) - 1 do
  begin
    thread := aThreads[i];
    if thread <> nil then
    begin
      if thread.ID = aId then
        result := thread;
    end
    else
      break;
  end;
end;

function AddThreadInfo(aId: TThreadId; aThreads: TThreadInfoDynArray): TThreadInfo;
var
  i: Integer;
begin
  result := TThreadInfo.Create(aId);
  for i := 0 to Length(aThreads) - 1 do
    if nil = aThreads[i] then
    begin
      aThreads[i] := result;
      break;
    end;
end;

procedure ClearThreads(aThreads: TThreadInfoDynArray);
var
  i: Integer;
begin
  for i := 0 to Length(aThreads) - 1 do
    if aThreads[i] <> nil then
    begin
      aThreads[i].Free;
      aThreads[i] := nil;
    end;
end;

// Obtain thread info for current thread;
// create new thread info & add it to Threads array if not found;
function ObtainThreadInfo: TThreadInfo;
var
  id: TThreadID;
begin
  EnterCriticalsection(ThreadsLocker);
  id := GetCurrentThreadId;
  result := FindThreadInfoById(id, Threads);
  if result = nil then
    result := AddThreadInfo(id, Threads);
  LeaveCriticalsection(ThreadsLocker);
end;

procedure instration_begin(const aUnit, aProc: string; aLine: Integer);
begin
  ObtainThreadInfo.Begin_(aUnit, aProc, aLine);
end;

procedure instration_end(const aUnit, aProc: string; aLine: Integer);
begin
  ObtainThreadInfo.End_(aUnit, aProc, aLine);
end;

function GetLogFileName: string;
begin
  result := GetModuleName(HINSTANCE) + '_prof_' + FormatDateTime('yyyy-mm-dd_hh-nn-ss', Now);
end;

{ TThreadInfo }

procedure TThreadInfo.PrepareTimerStack;
var
  i: Integer;
begin
  SetLength(TimerStack, MaxDepth);
  for i := 0 to Length(TimerStack) - 1 do
    Timer.Clear(TimerStack[i]);
end;

procedure TThreadInfo.WriteLog(s: string);
begin
  s := GetTimeStr + ': ' + GetIndent + s + LineEnding;
  Log.Write(s[1], Length(s));
end;

procedure TThreadInfo.PushTimer;
begin
  Timer.Clear(TimerStack[Level]);
  Timer.Start(TimerStack[Level]);
end;

function TThreadInfo.PopTimer: Extended;
begin
  Timer.Stop(TimerStack[Level]);
  result := Timer.Elapsed(TimerStack[Level]);
end;

function TThreadInfo.GetIndent: string;
var
  i: Integer;
begin
  SetLength(result, Level);
  for i := 1 to Level do
    result[i] := ' ';
end;

function TThreadInfo.GetTimeStr: string;
begin
  result := FormatDateTime('hh:mm:ss.zzz', Now);
end;

constructor TThreadInfo.Create(aID: TThreadID);
var
  logFilePath: string;
begin
  inherited Create;
  ID := aID;
  Timer := TEpikTimer.Create(nil);
  PrepareTimerStack;
  logFilePath := GetLogFileName + '.' + IntToStr(aID);
  WriteLN(logFilePath);
  Log := TFileStream.Create(logFilePath, fmCreate or fmOpenWrite);
end;

procedure TThreadInfo.Begin_(const aUnit, aProc: string; aLine: Integer);
begin
  Inc(Level);
  WriteLog('+' + aProc + ' ' + aUnit + ' L#' + IntToStr(aLine));
  PushTimer;
end;

procedure TThreadInfo.End_(const aUnit, aProc: string; aLine: Integer);
var
  time: Extended;
begin
  time := PopTimer;
  WriteLog('-' + aProc + ' ' + aUnit + ' ' + ' L#' + IntToStr(aLine)
    + ' time: ' + FormatFloat('0.000000', time) + ' sec');
  Dec(Level);
end;

destructor TThreadInfo.Destroy;
begin
  WriteLN('Now releasing thread info: ' + IntToStr(ID));
  Timer.Free;
  Log.Free;
  inherited Destroy;
end;

initialization
  InitCriticalSection(ThreadsLocker);
  SetLength(Threads, MaxThreads);
  ClearThreads(Threads);
finalization
  ClearThreads(Threads);
  DoneCriticalsection(ThreadsLocker);
end.

